import argparse
from torchvision import datasets, transforms
import torch.optim as optim
from DNN import *
from vat import *
import os
import numpy as np

from collections import namedtuple
from typing import List, Tuple, Any, Dict

labeled_batch_size = 32
test_batch_size = 100
unlabeled_batch_size = 128
num_labeled = 1000
num_iter_per_epoch = 400
eval_freq = 5
lr = 0.001
num_valid = 1000

parser = argparse.ArgumentParser()
parser.add_argument('--datapath', required=True, help='path to dataset')
parser.add_argument('--use_cuda', type=bool, default=True)
parser.add_argument('--num_epochs', type=int, default=200)
parser.add_argument('--epsilon', type=float, default=2.5)
parser.add_argument('--epoch_decay_start', type=int, default=160)

opt = parser.parse_args()

# test model return the error rate given the labeled data loader
def test_model(model, test_loader):
     error = 0
     total = 0
     model.eval()
     with torch.set_grad_enabled(False):
         for (inputs, yreal) in (test_loader):
             inputs, yreal = inputs.cuda(), yreal.cuda()
             ypred = model(inputs)
             _, predicted = torch.max(ypred.data, 1)

             total += yreal.size(0)
             error += (predicted != yreal).sum().item()

     test_error = 100 * error / total
     return test_error


# load the dataset
train_loader = torch.utils.data.DataLoader(
                            datasets.SVHN(root=opt.datapath, split='train', download=True,
                            transform=transforms.Compose([
                            transforms.ToTensor(),
                            transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))])),
                            batch_size=labeled_batch_size, shuffle=True)

test_loader = torch.utils.data.DataLoader(
                            datasets.SVHN(root=opt.datapath,  split='test', download=True,
                            transform=transforms.Compose([
                            transforms.ToTensor(),
                            transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)) ])),
                            batch_size=test_batch_size, shuffle=True)

train_data = []
train_target = []

for (data, target) in train_loader:
    train_data.append(data)
    train_target.append(target)

# divide the dataset into labeled, unlabaled dataset considering also validation dataset
train_data = torch.cat(train_data, dim=0)
train_target = torch.cat(train_target, dim=0)

# extract validation daa first
train_data, train_target = train_data[num_valid:,], train_target[num_valid:,]

labeled_train_data, labeled_train_target = train_data[:num_labeled,], train_target[:num_labeled,]
unlabeled_train_data = train_data[num_labeled:,]
print("Split training set into: {} unlabeled and {} labeled samples (validation dataset size is {})".format(unlabeled_train_data.size()[0], labeled_train_data.size()[0], num_valid))


# create the DNN and optimizer
model = ConvLarge()
model = model.cuda()
optimizer = optim.Adam(model.parameters(), lr=lr)
criterion = nn.CrossEntropyLoss()
vat_criterion = torch.nn.KLDivLoss(reduction = "batchmean") # use (reduction = "batchmean")

for epoch in range(opt.num_epochs):

    # decaying lr in last 40000 iterations
    if epoch > opt.epoch_decay_start:
        optimizer.lr = (opt.num_epochs - epoch) * lr / (opt.num_epochs - opt.epoch_decay_start)
        optimizer.betas = (0.5, 0.999)

    for i in range(num_iter_per_epoch):
        # get random data samples from labeled and unlabeled dataset
        batch_indices = torch.LongTensor(np.random.choice(len(labeled_train_data), labeled_batch_size, replace=False))
        xl = labeled_train_data[batch_indices]
        yl = labeled_train_target[batch_indices]
        batch_indices_star = torch.LongTensor(np.random.choice(len(train_data), unlabeled_batch_size, replace=False))
        x_star = train_data[batch_indices_star]
        xl, yl, x_star = xl.cuda(), yl.cuda(), x_star.cuda()

        model.train()

        with (torch.set_grad_enabled(True)):
            y_pred = model(xl)
            ce_loss = criterion(y_pred, yl)

            y_star = model(x_star)
            y_star = y_star.cuda()
            # calculate vat loss and add it to the cross entropy loss
            v_loss = vat_loss(model, vat_criterion, x_star, y_star, eps=opt.epsilon)
            loss = v_loss + ce_loss

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

        if i % 100 == 0:
            log_text = "Epoch: %d iter: %d, VAT loss: %.5f, CE loss: %.4f" %(epoch, i, v_loss, ce_loss)
            print(log_text)
            with open('epoch_logs.txt', 'a+') as f:
                f.write(log_text + '\n')
            
    if epoch % eval_freq == 0 or epoch + 1 == opt.num_epochs:
        test_error = test_model(model, test_loader)        
        log_text = "Epoch: %d, test error %.2f" %(epoch, test_error)
        print(log_text)
        with open('epoch_logs.txt', 'a+') as f:
            f.write(log_text + '\n')


